from typing import Type


class DependencyNotFoundException(Exception):
    def __init__(self, cls: Type, path: str = ""):
        self.cls = cls
        self.path = path
        super().__init__(f"Dependency {cls.__name__} not found while looking for {path}")
